#include <stdio.h>
#include <varargs.h>

/****************************************************************************\
  Hoopy form - by mirk@cs.warwick.ac.uk
  hform is similar to sprintf except it returns a pointer to a static buffer.
\****************************************************************************/

#define LINELEN 1024		/* Length of buffer */

char *hform (va_alist)
  va_dcl
{
  va_list pvar;
  char *fmt_string;
  static char result[LINELEN];

  va_start (pvar);
  fmt_string = va_arg (pvar, char*);
  (void) vsprintf (result, fmt_string, pvar);
  va_end (pvar);
  return (result);
}


/****************************************************************************\
  Hiscore functions. By max@cs.warwick.ac.uk
\****************************************************************************/

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include "score.h"

/*--------------------------------------------------------------------------*\
     Opens score for reading.

 Returns:  -3     - Could not open file that was there
 	   -2     - Could not open file descriptor for creation
 	   -1     - High score file access time out (lock file exists)
			perror may be called for system error.
	   otherwise
	          - file descriptor
\*--------------------------------------------------------------------------*/
int openscore ( file, timeout, flag )
char *file;			/* Score file name */
int timeout;			/* Timeout in seconds */
int flag;			/* 1 if we are to create file */
{
    int fd,
        lv,
        time = 0;

    /* Open and try to lock score file */
    if ( flag == CREATESCORE ) {
	if ( ( fd = open ( file, O_RDWR | O_CREAT, 0666 ) ) == -1 )
	    return ( -3 );
	if ( fchmod ( fd, 0666 ) == -1 )
	    return ( -3 );
    } else {
	if ( ( fd = open ( file, O_RDWR ) ) == -1 )
	    return ( -2 );
    }

    while (    ( lv = lockf ( fd, F_TLOCK, 0 ) == -1 )
	    && ( time <= timeout ) ) {
	(void) sleep ( 1 );
	time++;
    }
    if ( lv == -1 )
	return ( -1 );
    errno = 0;
    return ( fd );
}

/*--------------------------------------------------------------------------*\
    Closes score table
    Returns: -1 - error
	      0 - ok
\*--------------------------------------------------------------------------*/
int closescore ( fd )
int fd;
{
    if ( lockf ( fd, F_ULOCK, 0 ) == -1 ) {
	(void) close ( fd );
	return ( -1 );
    }
    (void) close ( fd );
    return ( 0 );
}

/****************************************************************************\
        Returns next signed integer in string or def if non.
	Places in len the length of string read.
\****************************************************************************/
int width ( point, def, len )
char *point;
int def;
int *len;
{
    int span;
    static char buff[WDBUFF];
    span = strspn ( point, "-0123456789" );
    *len = span;
    if ( span != 0 ) {
	bzero ( buff, sizeof ( buff ) );
	(void) strncpy ( buff, point, span );
	return ( atoi ( buff ) );
    } else
	return ( def );
}

/****************************************************************************\
  Clean score table, and initialise.

  0	- Ok intialiased
  -1	- timeout
  -2    - not around/open error
\****************************************************************************/
int newscore ( file, timeout )
char *file;
{
    int fd, i;
    SCORE zero;

    strcpy ( zero.name, "" );
    strcpy ( zero.info, "" );
    zero.score = (long) 0;
    zero.level = (long) 0;

    if ( ( fd = openscore ( file, timeout, CREATESCORE ) ) < 0 )
	return ( fd );

    (void) lseek ( fd, (off_t) 0, L_SET );
    for ( i = 0; i < STSIZE; i++ )
	(void) write ( fd, (char *) & zero, sizeof ( zero ) );

    return ( closescore ( fd ) );
}

/*--------------------------------------------------------------------------*\
  Adds score into table.

  Returns:
	        -1 Cannot write score for some reason
		 0 Score not written
		 Otherwise score ranking
\*--------------------------------------------------------------------------*/
int addscore ( fd, name, score, level, info )
char *name, *info;
int score, level, fd;
{
    int ts, i, found;
    SCORE *table;

    ts = sizeof ( SCORE ) * STSIZE;

    if ( ( table = (SCORE *) malloc ( ts ) ) == NULL ) {
	return ( -1 );
    }

    (void) lseek ( fd, (off_t) 0, L_SET );
				/* Read in table */
    if ( read ( fd, (char *) table, ts ) != ts ) {
	free ( table );
	return ( -1 );
    }

				/* Find position where score can go */
				/* -1 indicates nowhere */
    found = -1;
    for ( i = 0; i < STSIZE; i++ )
	if ( score >= table[i].score ) {
	    found = i;
	    break;
	}

    if ( found != -1 ) {	/* Insert score by moving things down */
	for ( i = (STSIZE-1); i >= found; i-- )
	    bcopy ( (char *) & table[i-1],
		    (char *) & table[i],
		    sizeof ( SCORE ) );
	bzero ( (char *) & table[found], sizeof ( SCORE ) );
	table[found].score = (long) score;
	table[found].level = (long) level;
	(void) strncpy ( table[found].name, name, NAMELEN );
	(void) strncpy ( table[found].info, info, INFOLEN );
	(void) lseek ( fd, (off_t) 0, L_SET );
	if ( write ( fd, (char *) table, ts ) != ts ) {
	    free ( table );
	    return ( -1 );
	}
    }

    free ( table );
    if ( found == -1 )
	return ( 0 );
    else
	return ( found + 1 );
}

/*--------------------------------------------------------------------------*\
    Returns score table in format specified.
    Name = %n
    Score = %s
    Level = %l
    Info  = %i

 	Notice: %20i means right justified in 20 spaces
                %-20i means left justified in 20 spaces
    Returns null if error.
\*--------------------------------------------------------------------------*/
char *getscore ( fd, num, format )
int fd, num;
char *format;
{
    off_t pos;
    SCORE *table;
    char *point;
    static char buffer[GSBUFFER];
    static char formbuf[GSBUFFER];
    static char realform[GSBUFFER];

    if ( num < 1 || num > STSIZE )
	return ( NULL );

				/* Confuses strtok later if line */
				/* starts with a %               */
    if ( format[0] != '%' )
	(void) strcpy ( formbuf, format );
    else {
	formbuf[0] = ' ';
	formbuf[1] = '\0';
	(void) strcat ( formbuf, format );
    }

    if ( ( table = (SCORE *) malloc ( sizeof ( SCORE ) ) ) == NULL )
	return ( NULL );

				/* Seek to correct position */
    pos = sizeof ( SCORE ) * (num - 1);
    if ( lseek ( fd, pos, L_SET ) != pos ) {
	free ( table );
	return ( NULL );
    }
				/* Read in table entry  */
    if ( read ( fd, (char *) table, sizeof ( SCORE ) ) != sizeof ( SCORE ) ) {
	free ( table );
	return ( NULL );
    }

    bzero ( buffer, sizeof ( buffer ) );

    if ( ( point = strtok ( formbuf, "%" ) ) != NULL ) {
	(void) strcat ( buffer, point );
	while ( ( point = strtok ( NULL, "%" ) ) != NULL ) {
	    int length = 0, formwidth, actual;

	    formwidth = width ( point, 10, &length );
	    point += length;
	    switch ( point[0] ) {
	    case 'n':
		(void) strcpy ( realform, hform ( "%%%ds", formwidth ) );
		(void) strcat ( buffer, hform ( realform, table->name ) );
		point++;
		break;
	    case 's':
		(void) strcpy ( realform, hform ( "%%%dld", formwidth ) );
		(void) strcat ( buffer, hform ( realform, table->score ) );
		point++;
		break;
	    case 'l':
		(void) strcpy ( realform, hform ( "%%%dld", formwidth ) );
		(void) strcat ( buffer, hform ( realform, table->level ) );
		point++;
		break;
	    case 'i':
		(void) strcpy ( realform, hform ( "%%%ds", formwidth ) );
		(void) strcat ( buffer, hform ( realform, table->info ) );
		point++;
		break;
	    }
	    (void) strcat ( buffer, point );
	}
    } else
	(void) strcat ( buffer, format );
				/* End of getting things */
    free ( table );
    return ( buffer );
}

